/************************************************************************************
 **                                                                                **
 **                         IPS Utility for Windows9x                              **
 **                                                                                **
 **                                                                                **
 **                          COPYRIGHT (c) 1999 M.K.S                              **
 **                                                                                **
 **                                                                                **
 **  < About IPS Patching from WaveMaster's doc >                                  **
 **                                                                                **
 **           Read the first 5 bytes of the .IPS file (the header)                 **
 **           If header != "PATCH" [0x50,0x41,0x54,0x43,0x48], ***END PROGRAM***   **
 **                                                                                **
 **  loop ---> Read the next 3 bytes from the .IPS (let's say, RST)                **
 **   |                                                                            **
 **   |        If RST == "EOF"[0x45,0x4f,0x46] then *** END PROGRAM ***            **
 **   |                                                                            **
 **   |        If RST != "EOF" then RST represents the ADDRESS of source file      **
 **   |        that the next chunk of .IPS data will be placed at.                 **
 **   |                                                                            **
 **   |        Read the next word (let's say XY)                                   **
 **   |                                                                            **
 **   |        If XY != 0 then xy is the number of bytes to read                   **
 **   |        from the .IPS file.  Read those bytes from the .IPS file, and       **
 **   |        copy the bytes to the address (rst) of the source file              **
 **   +------- (Return to top of loop)                                             **
 **   |                                                                            **
 **   |        If XY == 0 then read the next word (let's say "AB"), then read      **
 **   |        the next byte (let's say "C")                                       **
 **   |        Copy byte C, AB times to the source file, at address RST            **
 **   +------- (return to top of loop)                                             **
 **                                                                                **
 **  < History >                                                                   **
 **    version 0.10 [ Feb 9,1999 ] - Initial release                               **
 **                                                                                **
 ************************************************************************************/

#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <stdio.h>
#include "ips.h"
#include "myfunc.h"

static void PostProgressMessage(HWND hwndParent,UINT uMsg,UINT uNew,LPUINT lpuOld,LPTSTR lpStatus)
{
	if(!hwndParent) return;

	if(lpuOld){
		if(*lpuOld + 5 <= uNew){
			SendMessage(hwndParent,uMsg,(WPARAM)uNew,(LPARAM)lpStatus);
			*lpuOld = uNew;
		}
	}
	else
		SendMessage(hwndParent,uMsg,(WPARAM)uNew,(LPARAM)lpStatus);
}

/*****************

	IPSGW

******************/
BOOL ips_IsIPSFile(LPCTSTR lpIPSFile)
{
	FILE *fpIPS;
	BYTE header[5];
	BOOL blResult=TRUE;

	if(!lpIPSFile) return FALSE;

	// Open ips file
	//---------------
	if((fpIPS=fopen(lpIPSFile,"rb"))==NULL) return FALSE;

	// Check if .ips file has "PATCH" header 
	//---------------------------------------
	fread(header,1,5,fpIPS);
	if(memcmp(header,"\x50\x41\x54\x43\x48",5)) blResult=FALSE;

	fclose(fpIPS);

	return blResult;
}

//-------------
//@IPSpb`
//-------------
BOOL ips_Patch(LPCTSTR lpSrcFile,LPCTSTR lpIPSFile,LPTSTR lpLogBuf,HWND hwndParent,UINT uMsg)
{
	FILE *fpIPS;
	FILE *fpSrc;
	BYTE header[5];
	union 
	{
		unsigned char byte[4];
		unsigned long int dword;
	}RST;
	union
	{
		unsigned char byte[2];
		unsigned short int word;
	}XY;
	union
	{
		unsigned char byte[2];
		unsigned short int word;
	}AB;
	unsigned char c;
	register int i;
	BOOL blResult;
	long dwIPSFileSize,dwFileSize;
	UINT uPercentage;

	if(!lpSrcFile || !lpIPSFile) return FALSE;

	//
	//--------------
	uPercentage = 0;
	blResult = FALSE;

	PostProgressMessage(hwndParent, uMsg, 0, NULL, "Getting file information...");

	// Open ips file
	//---------------
	if((fpIPS = fopen(lpIPSFile, "rb")) == NULL){
		if(lpLogBuf)
			wsprintf(lpLogBuf, "%s could not be opened.", lpIPSFile);
		return FALSE;
	}

	// Get ips file size
	//----------------------
	fseek(fpIPS, 0, SEEK_END);
	dwIPSFileSize = ftell(fpIPS);
	fseek(fpIPS, 0, SEEK_SET);

	// Check if .ips file has "PATCH" header 
	//---------------------------------------
	fread(header, 1, 5, fpIPS);
	if(memcmp(header, "\x50\x41\x54\x43\x48", 5)){
		if(lpLogBuf)
			wsprintf(lpLogBuf, "%s is not an IPS file.", lpIPSFile);
		fclose(fpIPS);
		return FALSE;
	}

	// Open Source file
	//------------------
	if((fpSrc = fopen(lpSrcFile, "r+b")) == NULL){
		if(lpLogBuf)
			wsprintf(lpLogBuf, "%s could not be opened in write mode.", lpSrcFile);
		fclose(fpIPS);
		return FALSE;
	}

	// Get source file size
	//----------------------
	fseek(fpSrc, 0, SEEK_END);
	dwFileSize = ftell(fpSrc);
	fseek(fpSrc, 0, SEEK_SET);

	//
	//
	if(dwIPSFileSize==-1L
		|| dwFileSize == -1L){
		if(lpLogBuf)
			wsprintf(lpLogBuf,"Couldn't get file size.");
		goto error;
	}

	// Set initial Value
	//-------------------
	RST.byte[3]=0x00;

	// Patching
	//----------
	for(;;){
		
		if(feof(fpIPS)){
			if(lpLogBuf)
				wsprintf(lpLogBuf,"EOF not found.");
			break;
		}

		// Read chunk
		//------------
		
		/*
		// old version
		for(i=2;i>=0;i--)
			fread(RST.byte+i,1,1,fpIPS);
		*/

		fread(RST.byte, 3, 1, fpIPS);
		{
			BYTE bBuf;
			bBuf = RST.byte[2];
			RST.byte[2] = RST.byte[0];
			RST.byte[0] = bBuf;
		}

		//
		PostProgressMessage(hwndParent, uMsg, 100 * ftell(fpIPS) / dwIPSFileSize, &uPercentage, "Processing patch...");

		// Check "EOF"
		//------------
		if(!memcmp(RST.byte, "\x46\x4f\x45", 3)){
			//
			PostProgressMessage(hwndParent, uMsg, 100, NULL, "Complete");
			if(lpLogBuf){
				wsprintf(lpLogBuf, "%s was patched successfully.", lpSrcFile);
			}
			blResult = TRUE;
			break;
		}

		// Set source file pointer
		//-------------------------
		/*
		\[Xt@CƉσt@CTCYႢAɂ̑̃f[^̃`Nɕꂽ
		IPSt@C̏ꍇA\[Xt@C̃ItZbg𒴂ꍇ̂ł̃G[`FbŃA
		͂ȂB
		if( dwFileSize>(long)RST.dword ){
			fseek(fpSrc,RST.dword,SEEK_SET);
		}
		else{
			if(lpLogBuf)
				wsprintf(lpLogBuf,"IPS file offset is invalid.");
			break;
		}
		*/
		fseek(fpSrc, RST.dword, SEEK_SET);

		// Read XY
		//---------
		/*
		fread(XY.byte+1,1,1,fpIPS);
		fread(XY.byte,1,1,fpIPS);
		*/
		fread(XY.byte, 2, 1, fpIPS);
		{
			BYTE bBuf;
			bBuf = XY.byte[1];
			XY.byte[1] = XY.byte[0];
			XY.byte[0] = bBuf;
		}


		if(XY.word){
			LPBYTE lpbBuf;
			// Copy data
            //-----------
			lpbBuf=malloc(XY.word);
			if(!lpbBuf){
				if(lpLogBuf)
					wsprintf(lpLogBuf,"Memory allocation failed.");
				break;
			}
		    fread(lpbBuf, XY.word, 1, fpIPS);
			fwrite(lpbBuf, XY.word, 1, fpSrc);
			free(lpbBuf);
		}
		else{
			// Read AB
			//---------
			fread(AB.byte + 1, 1, 1, fpIPS);
			fread(AB.byte, 1, 1, fpIPS);
			// Read c
			//--------
			fread(&c, 1, 1, fpIPS);
			// Copy Data
			//-----------
			for(i = 0;i < AB.word;i++)
				fwrite(&c, 1, 1, fpSrc);
		}
	}

error:

	// Close files
	//-------------
	fclose(fpIPS);
	fclose(fpSrc);

	return blResult;
}

//-------------------
//@IPSt@C쐬
//-------------------

//@obt@̂ꂼ̃oCgׂēׂ֐
static BOOL IsAllEqual(LPBYTE pbBuf,DWORD dwSize)
{
	BOOL blRet;
	DWORD N;

	blRet=TRUE;
	for(N=0;N<dwSize-1;N++){
		if(pbBuf[N]!=pbBuf[N+1]){
			blRet=FALSE;
			break;
		}
	}
	return blRet;
}

//
//@IPSwb_[
//
static BOOL IPSHeaderWrite(HANDLE hFile)
{
	BYTE bHead[]="\x50\x41\x54\x43\x48";
	DWORD dwBytes;

	if(!WriteFile(hFile,bHead,5,&dwBytes,NULL) || dwBytes!=5 ) return FALSE;

	return TRUE;
}

//
//@IPS`N
//
static BOOL IPSChunkWrite(DWORD dwOffset, DWORD dwSize, LPBYTE pbBuf, HANDLE hFile)
{
	int i;
	DWORD B;
	WORD wChunks;
	DWORD dwWriteSize;
	
	// IPSt@ĆA0xFFFFFF(16777215) oCg ȏ̃ItZbgƂłȂB
	if(dwOffset > 0xFFFFFF)
		return FALSE;

	wChunks=(WORD)(dwSize>>16);
	if(dwSize>3 && IsAllEqual(pbBuf,dwSize)){
		// Offset(3) | 00 00 | Times(2) | data(1)
		do{
			if(wChunks){
				dwWriteSize=0xFFFF;
				dwSize-=0xFFFF;
			}
			else{
				dwWriteSize=dwSize;
				//dwSize-=(WORD)dwSize;
			}

			// dwOffset
			// dwWriteSize
			// pbBuf
			for(i=2;i>=0;i--)
				WriteFile(hFile,(BYTE*)&dwOffset+i,1,&B,NULL);
			WriteFile(hFile,"\x00\x00",2,&B,NULL);
			for(i=1;i>=0;i--)
				WriteFile(hFile,(BYTE*)&dwWriteSize+i,1,&B,NULL);
			WriteFile(hFile,pbBuf,1,&B,NULL);

			dwOffset+=(WORD)dwWriteSize;

		}while(wChunks--);
	}
	else{
		// Offset(3) | size(2) | data .. data(size)
		do{
			if(wChunks){
				dwWriteSize = 0xFFFF;
				dwSize -= 0xFFFF;
			}
			else{
				dwWriteSize = dwSize;
				//dwSize-=(WORD)dwSize;
			}

			for(i = 2;i >= 0;i--)
				WriteFile(hFile, (BYTE*)&dwOffset + i, 1, &B, NULL);
			for(i = 1;i >= 0;i--)			
				WriteFile(hFile, (BYTE*)&dwWriteSize + i, 1, &B, NULL);
			WriteFile(hFile, pbBuf, (WORD)dwWriteSize, &B, NULL);

			dwOffset+=(WORD)dwWriteSize;
			pbBuf += dwWriteSize;

		}while(wChunks--);
	}
	return TRUE;
}

//
//@EOF
//
static BOOL IPSEofWrite(HANDLE hFile)
{
	BYTE bEof[]="\x45\x4f\x46";
	DWORD dwBytes;

	if(!WriteFile(hFile,bEof,3,&dwBytes,NULL)) return FALSE;

	return TRUE;
}

//
//@쐬̂߂̃C[`
//
BOOL ips_Create(LPCTSTR lpFileOrg,LPCTSTR lpFileChg,LPCTSTR lpFileIPS,LPTSTR lpLogBuf,HWND hwndParent,UINT uMsg)
{
	BYTE *pbOrg;
	BYTE *pbChg;
	DWORD dwOfs;
	BYTE *pbChgTmp;
	FILEMEMORYINFO sFMInfoOrg;
	FILEMEMORYINFO sFMInfoChg;
	BOOL blRet;
	HANDLE hFileIPS;
	DWORD dwFileSize;
	UINT uPercentage;

	blRet=TRUE;

	PostProgressMessage(hwndParent,uMsg,0,NULL,"Getting file information...");

	// Open unchanged file
	//---------------------
	if(!AllocateFileMemory(lpFileOrg,&sFMInfoOrg)){
		if(lpLogBuf) wsprintf(lpLogBuf,"%s could not be opened.",lpFileOrg);
		return FALSE;
	}

	// Open changed file
	//---------------------
	if(!AllocateFileMemory(lpFileChg,&sFMInfoChg)){
		if(lpLogBuf) wsprintf(lpLogBuf,"%s could not be opened.",lpFileChg);
		FreeFileMemory(&sFMInfoOrg);
		return FALSE;
	}

	if(sFMInfoOrg.dwFileSize>sFMInfoChg.dwFileSize){
		if(lpLogBuf) wsprintf(lpLogBuf,"Cannot create IPS because the size of the file after modification is smaller than the size before modification.");
		blRet=FALSE;
		goto EXIT_FUNCTION;
	}

	//
	dwFileSize = sFMInfoOrg.dwFileSize;
	uPercentage = 0;

	hFileIPS=CreateFile(lpFileIPS,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
	if(hFileIPS==INVALID_HANDLE_VALUE){
		if(lpLogBuf) wsprintf(lpLogBuf,"Could not create new file.",lpFileIPS);
		blRet=FALSE;
		goto EXIT_FUNCTION;
	}

	IPSHeaderWrite(hFileIPS);

	pbOrg=sFMInfoOrg.lpPointer;
	pbChg=sFMInfoChg.lpPointer;
	pbChgTmp=NULL;
	dwOfs=0;
	for(;dwOfs<sFMInfoOrg.dwFileSize;pbOrg++,pbChg++,dwOfs++){
		if(pbOrg[0]==pbChg[0] && pbChgTmp){
			//
			PostProgressMessage(hwndParent,uMsg,100*dwOfs/dwFileSize,&uPercentage,"Creating...");
			//
			if(!IPSChunkWrite(dwOfs-(pbChg-pbChgTmp),pbChg-pbChgTmp,pbChgTmp,hFileIPS)){
				blRet=FALSE;
				if(lpLogBuf)
					wsprintf(lpLogBuf,"An error occurred during creation.",lpFileIPS);
				goto EXIT_FUNCTION;
			}

			pbChgTmp=NULL;
		}
		else if(pbOrg[0]!=pbChg[0] && !pbChgTmp){
			pbChgTmp=pbChg;
		}
	}/* for */

	//Ō̃oCgƈĂꍇ
	if(pbChgTmp)
		IPSChunkWrite(dwOfs - (pbChg - pbChgTmp), pbChg - pbChgTmp, pbChgTmp, hFileIPS);

	//TCY𒴂
	if(sFMInfoOrg.dwFileSize < sFMInfoChg.dwFileSize) {
		IPSChunkWrite(dwOfs , sFMInfoChg.dwFileSize - sFMInfoOrg.dwFileSize , pbChg , hFileIPS);
	}

	IPSEofWrite(hFileIPS);

	//
	PostProgressMessage(hwndParent,uMsg,100,NULL,"Complete");

	if(lpLogBuf) wsprintf(lpLogBuf,"IPS file %s was created.",lpFileIPS);

EXIT_FUNCTION:
	FreeFileMemory(&sFMInfoOrg);
	FreeFileMemory(&sFMInfoChg);
	CloseHandle(hFileIPS);

	return blRet;
}
