#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <string.h>
#include "winips.h"
#include "ips.h"
#include "ipslib.h"
#include "myfunc.h"

BOOL IsIPSFile(LPCTSTR lpFilePath)
{
	return ips_IsIPSFile(lpFilePath);
}

BOOL CreateIPSFile(LPIPSLIBPARAMS lpParams)
{
	LPTSTR lpOrgPath = lpParams->szFilePath1;
	LPTSTR lpChgPath = lpParams->szFilePath2;
	HWND hwndParent = lpParams->hWnd;
	BOOL blYesNo;
	TCHAR szLog[LOG_BUFFER_SIZE];
	TCHAR cIPSBuf[MAX_PATH];
	TCHAR szMsg[MAX_PATH*3+100];

	SendMessage(hwndParent,WM_IPSLIB_BEGIN,0,0);

	if(!CheckFileExistance(lpOrgPath)){
		wsprintf(szMsg,"%s doesn't exist.",lpOrgPath);
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	if(!CheckFileExistance(lpChgPath)){
		wsprintf(szMsg,"%s doesn't exist.",lpChgPath);
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	if(!lstrcmp(lpOrgPath,lpChgPath)){
		wsprintf(szMsg,"The two file names are the same.");
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	ChangeFileExtention(lpChgPath,cIPSBuf,MAX_PATH*sizeof(TCHAR),"ips");

	wsprintf(szMsg,"Unchanged file\n%s\nis changed to the file\n%s\nby patching it with the IPS file\n%s\nwhich will be created.\nIs this OK?",lpOrgPath,lpChgPath,cIPSBuf);
	SendMessage(hwndParent,WM_IPSLIB_CONFIRM,(WPARAM)&blYesNo,(LPARAM)szMsg);
	if(!blYesNo){
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)NULL);
		return FALSE;
	}

	if(CheckFileExistance(cIPSBuf)){
		wsprintf(szMsg,"%s already exists.\nCan it be overwritten?",cIPSBuf);
		SendMessage(hwndParent,WM_IPSLIB_CONFIRM,(WPARAM)&blYesNo,(LPARAM)szMsg);
		if(!blYesNo){
			SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)NULL);
			return FALSE;
		}
	}

	if(ips_Create(lpOrgPath,lpChgPath,cIPSBuf,szLog,hwndParent,WM_IPSLIB_PROGRESS)){
		SendMessage(hwndParent,WM_IPSLIB_FINISH,TRUE,(LPARAM)szLog);
	}
	else{
		DeleteFile(cIPSBuf);
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szLog);
	}
	return TRUE;
}

BOOL IPSPatch(LPIPSLIBPARAMS lpParams)
{
	LPTSTR lpIPSPath = lpParams->szFilePath1;
	LPTSTR lpSrcPath = lpParams->szFilePath2;
	HWND hwndParent = lpParams->hWnd;
	BOOL blYesNo;
	TCHAR szLog[LOG_BUFFER_SIZE];
	TCHAR cNewBuf[MAX_PATH];
	TCHAR cSrcFileExt[MAX_PATH];
	TCHAR szMsg[MAX_PATH*3+100];

	SendMessage(hwndParent,WM_IPSLIB_BEGIN,0,0);

	if(!CheckFileExistance(lpIPSPath)){
		wsprintf(szMsg,"%s doesn't exist.",lpIPSPath);
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	if(!CheckFileExistance(lpSrcPath)){
		wsprintf(szMsg,"%s doesn't exist.",lpSrcPath);
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	if(!ips_IsIPSFile(lpIPSPath)){
		wsprintf(szMsg,"%s is not an IPS file.",lpIPSPath);
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	if(!lstrcmp(lpIPSPath,lpSrcPath)){
		wsprintf(szMsg,"The two file names are the same.");
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	GetFileExtention(lpSrcPath,cSrcFileExt,sizeof(cSrcFileExt));
	ChangeFileExtention(lpIPSPath,cNewBuf,MAX_PATH,cSrcFileExt);

	wsprintf(szMsg,"File to be patched\n%s\nits copy\n%s\nis made, and the IPS file\n%s\npatches it.\nIs this OK?",lpSrcPath,cNewBuf,lpIPSPath);
	SendMessage(hwndParent,WM_IPSLIB_CONFIRM,(WPARAM)&blYesNo,(LPARAM)szMsg);
	if(!blYesNo){
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)NULL);
		return FALSE;
	}

	if(!_stricmp(cNewBuf,lpSrcPath)){
		wsprintf(szMsg,"Because the file to be patched and the copy file have the same name, it cannot be patched.");
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	if(CheckFileExistance(cNewBuf)){
		wsprintf(szMsg,"%s already exists.\nCan it be overwritten?",cNewBuf);
		SendMessage(hwndParent,WM_IPSLIB_CONFIRM,(WPARAM)&blYesNo,(LPARAM)szMsg);
		if(!blYesNo){
			SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)NULL);
			return FALSE;
		}

		SetFileAttributes(cNewBuf,FILE_ATTRIBUTE_NORMAL);
	}
	
	if(!CopyFile(lpSrcPath,cNewBuf,FALSE)){
		wsprintf(szMsg,"File could not be copied.");
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szMsg);
		return FALSE;
	}

	SetFileAttributes(cNewBuf,FILE_ATTRIBUTE_NORMAL);

	if(ips_Patch(cNewBuf,lpIPSPath,szLog,hwndParent,WM_IPSLIB_PROGRESS)){
		SendMessage(hwndParent,WM_IPSLIB_FINISH,TRUE,(LPARAM)szLog);
	}
	else{
		DeleteFile(cNewBuf);
		SendMessage(hwndParent,WM_IPSLIB_FINISH,FALSE,(LPARAM)szLog);
	}

	return TRUE;
}
